﻿/*
	BitmapPool.cs
	Copyright 2016, OSTRA / Takeshi Yoneki
*/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.IO.Compression;
using System.Text;
using System.Runtime.InteropServices;

namespace PictureSetViewLite {
	// 画像ファイルのキャッシュプール
	public class BitmapPool {
		public enum ContainerType {
			NONE, DIR, ZIP
		}
		const int POOL_COUNT = 3;
		public ContainerType containerType = ContainerType.NONE;
		public String filePath = "";

		// スレッドの簡易Actor
		Actor loaderAct = null;
		class LoaderMessage {
			public enum Method {
				NONE, FILES, GETNTH, FILL, GETCOUNT, GETINDEXOFFILES, CHANGESORT, CHANGEWITHSUBDIRECTORY
			}
			public Method method = Method.NONE;
			public int index = 0;
			public Bitmap bmp = null;
			public int count = 0;
			public bool alphabetical = false;
			public bool withSubdirectory = false;
			public String filePath = "";
		}

		void LoaderActFiles(bool s, bool a) {
			if (loaderAct != null) {
				loaderAct.Post(new LoaderMessage { method = LoaderMessage.Method.FILES, withSubdirectory = s, alphabetical = a });
			}
		}

		int LoaderActGetCount() {
			if (loaderAct != null) {
				return ((LoaderMessage)loaderAct.Send(new LoaderMessage { method = LoaderMessage.Method.GETCOUNT })).count;
			}
			return 0;
		}

		int LoaderActGetIndexOfFiles(String p) {
			if (loaderAct != null) {
				return ((LoaderMessage)loaderAct.Send(new LoaderMessage { method = LoaderMessage.Method.GETINDEXOFFILES, filePath = p })).index;
			}
			return 0;
		}

		void LoaderActChangeSort(bool a) {
			if (loaderAct != null) {
				loaderAct.Send(new LoaderMessage { method = LoaderMessage.Method.CHANGESORT, alphabetical = a });
			}
		}

		void LoaderActChangeWithSubdirectory(bool s, bool a) {
			if (loaderAct != null) {
				loaderAct.Send(new LoaderMessage { method = LoaderMessage.Method.CHANGEWITHSUBDIRECTORY, withSubdirectory = s, alphabetical = a });
			}
		}

		void LoaderActFill(int n) {
			if (loaderAct != null && IsEnableGet(n)) {
                loaderAct.StopRequest();
                loaderAct.Post(new LoaderMessage { method = LoaderMessage.Method.FILL, index = n });
			}
		}

		Bitmap LoaderActGetNth(int n) {
			if (loaderAct != null && IsEnableGet(n)) {
                loaderAct.StopRequest();
                return ((LoaderMessage)loaderAct.Send(new LoaderMessage { method = LoaderMessage.Method.GETNTH, index = n })).bmp;
			}
			return null;
		}

		void LoaderActJoin() {
			if (loaderAct != null) {
				loaderAct.Quit();
				loaderAct.Join();
				loaderAct = null;
			}
		}

		Actor imagesAct = null;
		class ImagesMessage {
			public enum Method {
				NONE, ADD, GETNTH, KEEP, CLEAR
			}
			public Method method = Method.NONE;
			public int index = 0;
			public Bitmap bmp = null;
			public List<int> keep = null;
		}

		void ImagesActClear() {
			if (imagesAct != null) {
				imagesAct.Send(new ImagesMessage { method = ImagesMessage.Method.CLEAR });
			}
		}

		Bitmap ImagesActGetNth(int n) {
			if (imagesAct != null) {
				return ((ImagesMessage)imagesAct.Send(new ImagesMessage { method = ImagesMessage.Method.GETNTH, index = n })).bmp;
			}
			return null;
		}

		void ImagesActAdd(int n, Bitmap bm) {
			if (imagesAct != null) {
				imagesAct.Post(new ImagesMessage { method = ImagesMessage.Method.ADD, index = n, bmp = bm });
			}
		}

		void ImagesActKeep(List<int> kp) {
			if (imagesAct != null) {
				imagesAct.Post(new ImagesMessage { method = ImagesMessage.Method.KEEP, keep = kp });
			}
		}

		void ImagesActJoin() {
			if (imagesAct != null) {
				imagesAct.Quit();
				imagesAct.Join();
				imagesAct = null;
			}
		}

		// コンストラクタ
		public BitmapPool(String inPath, bool withSubdirectory, bool alphabetical) {
            if (inPath.Length == 0) {
                return;
            }
            filePath = inPath;
            containerType = ContainerType.NONE;
            if (Directory.Exists(filePath)) {
                containerType = ContainerType.DIR;
            }
            else if (File.Exists(filePath)) {
                if (HasZipExt(filePath)) {
                    containerType = ContainerType.ZIP;
                }
                else {
                    containerType = ContainerType.DIR;
                    filePath = Path.GetDirectoryName(filePath);
                }
            }
            else {
                return;
            }

            // 最初はGetNthの度にスレッドを生成する大量消費型だったが、
            // Monitorクラスでメッセージングが可能なので1スレッド固定型に変更。
            // 1スレッドなら起動の軽量なTaskやThreadPoolである必然性もない。
            // TaskやThreadPoolは大量消費型では適切に動けないのでどのみち使う理由はない。
            // Actorの簡易実装に変更。
            loaderAct = new Actor(LoaderActorMain);
			imagesAct = new Actor(ImagesActorMain);
			// 生成したパスリストはloaderActにて保持する。
			LoaderActFiles(withSubdirectory, alphabetical);
        }

		public void Join() {
			LoaderActJoin();
			ImagesActJoin();
        }

        // ファイルパスから画像番号を得る
        public int GetIndexOfFiles(String path) {
			return LoaderActGetIndexOfFiles(path);
        }

        // ソート変更
        public void ChangeSort(bool alphabetical) {
			LoaderActChangeSort(alphabetical);
		}

		// サブディレクトリを含むかを変更
		// 再ソートも行う
		public void ChangeWithSubdirectory(bool withSubdirectory, bool alphabetical) {
			LoaderActChangeWithSubdirectory(withSubdirectory, alphabetical);
        }

        // 画像ファイルか
        public static bool HasImageExt(String name) {
            String nm = name.ToLower();
            return nm.EndsWith(".jpg") || nm.EndsWith(".jpeg") || nm.EndsWith(".png") || nm.EndsWith(".gif") || nm.EndsWith(".bmp") || nm.EndsWith(".tiff") || nm.EndsWith(".tif");
        }

        // ZIPファイルか
        static bool HasZipExt(String name) {
            String nm = name.ToLower();
            return nm.EndsWith(".zip") || nm.EndsWith(".cbz");
        }

        // ImagePoolインスタンスの抱えるディレクトリパスまたはZIPファイルパスが有効か
        public bool IsFileExist() {
            if (containerType == ContainerType.DIR && Directory.Exists(filePath)) {
                return true;
            }
            else if (containerType == ContainerType.ZIP && File.Exists(filePath)) {
                return true;
            }
            else {
                return false;
            }
        }

        // 画像ファイル数
        public int GetCount() {
			return LoaderActGetCount();
        }

		bool IsEnableGet(int n) {
			return n >= 0 && n < LoaderActGetCount() && IsFileExist();
		}

		// 画像キャッシュ充填。
		public void Fill(int n) {
			LoaderActFill(n);
		}

		// 指定の番号の画像ファイルのBitmapを得る
		public Bitmap GetNth(int n) {
			return LoaderActGetNth(n);
		}

		// 画像ファイルパスを再帰的に得る
		List<String> FetchFilePathList(String inPath, bool withSubdirectory) {
			List<String> files = new List<String>();
			foreach (String path in Directory.GetFiles(inPath)) {
				if (HasImageExt(path)) {
					files.Add(path);
					//Trace.WriteLine(path);
				}
			}
			if (withSubdirectory) {
				foreach (String path in Directory.GetDirectories(inPath)) {
					files.AddRange(FetchFilePathList(path, withSubdirectory));
				}
			}
			return files;
		}

		List<String> LoaderActorMain_Files(bool withSubdirectory) {
			List<String> files = new List<String>();
			if (containerType == ContainerType.DIR) {
				files = FetchFilePathList(filePath, withSubdirectory);
			}
			else if (containerType == ContainerType.ZIP) {
				using (var fs = new FileStream(filePath, FileMode.Open)) {
					using (var zip = new ZipArchive(fs, ZipArchiveMode.Read, false, Encoding.GetEncoding(932))) {
						foreach (ZipArchiveEntry e in zip.Entries) {
							if (HasImageExt(e.FullName)) {
								files.Add(e.FullName);
								//Trace.WriteLine(e.FullName);
							}
						}
					}
				}
			}
			return files;
		}

		// 関数内関数がまだないので
		List<String> LoaderActorMain_ChangeSort(List<String> files, bool alphabetical) {
			if (alphabetical) {
				files.Sort(StringComparer.OrdinalIgnoreCase);
			}
			else {
				files.Sort(new NaturalOrderStringComparer());
			}
			return files;
		}

		// 画像ファイルの読み込みは別スレッド
		// imagesに画像データを溜める。
		void LoaderActorMain(Object arg) {
            var act = (Actor)arg;
            FileStream fs = null;
            ZipArchive zip = null;
			List<String> files = new List<String>();
			Object o;
			while ((o = act.Receive()) != null) {
				var msg = (LoaderMessage)o;
				if (msg.method == LoaderMessage.Method.FILES) {
					files = LoaderActorMain_ChangeSort(LoaderActorMain_Files(msg.withSubdirectory), msg.alphabetical);
				}
				else if (msg.method == LoaderMessage.Method.GETCOUNT) {
					msg.count = files.Count;
					act.Reply(msg);
				}
				else if (msg.method == LoaderMessage.Method.GETINDEXOFFILES) {
					msg.index = 0;
					for (int i = 0; i < files.Count; i++) {
						if (String.Compare(files[i], msg.filePath, true) == 0) {
							msg.index = i;
							break;
						}
					}
					act.Reply(msg);
				}
				else if (msg.method == LoaderMessage.Method.CHANGESORT) {
					files = LoaderActorMain_ChangeSort(files, msg.alphabetical);
					ImagesActClear();
					act.Reply(null);
				}
				else if (msg.method == LoaderMessage.Method.CHANGEWITHSUBDIRECTORY) {
					if (containerType == ContainerType.DIR) {
						files = LoaderActorMain_ChangeSort(FetchFilePathList(filePath, msg.withSubdirectory), msg.alphabetical);
					}
					ImagesActClear();
					act.Reply(null);
				}
				else if (msg.method == LoaderMessage.Method.GETNTH || msg.method == LoaderMessage.Method.FILL) {
					var order = new List<int>();
					// 指定された番号の画像の前後もキャッシュとして読み込む
					// "only"時は指定された番号の画像のみ読み込む。
					// "fill"時は指定された番号の前後の画像も読み込む。
					order.Add(msg.index);
					if (msg.method == LoaderMessage.Method.FILL) {
						for (int i = 0, n = msg.index + 1; i < POOL_COUNT && n < files.Count; i++, n++) {
							order.Add(n);
						}
						for (int i = 0, n = msg.index - 1; i < POOL_COUNT && n >= 0; i++, n--) {
							order.Add(n);
						}
					}
					// Bitmap生成時にファイルのストリームを作って渡した場合、すぐCloseをすると適切に扱えない。
					// ファイルパスを渡せるので、そっちを使えということか。
					// ZipArchiveEntryのストリームはすぐCloseしても問題ない。
					// こっちは他の方法がないので、動いてくれないと途方に暮れる。
					// この2つのストリームで別の動きをしているのだが、MSDNでもドキュメント化されてそうもない。
					Bitmap bmp = null;
					foreach (int n in order) {
						if (msg.method == LoaderMessage.Method.FILL && act.IsStopRequested()) {
                        //if (msg.method == LoaderMessage.Method.FILL && act.HasCue()) {
                                break;
						}
						bmp = ImagesActGetNth(n);
						if (bmp != null) {
							continue;
						}
						Trace.WriteLine(files[n]);
						try {
							if (containerType == ContainerType.DIR) {
								if (File.Exists(files[n])) {
									bmp = new Bitmap(files[n]);
								}
							}
							else if (containerType == ContainerType.ZIP) {
								if (zip == null) {
									fs = new FileStream(filePath, FileMode.Open);
									zip = new ZipArchive(fs, ZipArchiveMode.Read, false, Encoding.GetEncoding(932));
								}
								var e = zip.GetEntry(files[n]);
								if (e != null) {
									var m = e.Open();
									bmp = new Bitmap(m);
									m.Close();
								}
							}
						}
						catch (Exception) {
						}
						ImagesActAdd(n, bmp);
					}
					if (msg.method == LoaderMessage.Method.GETNTH) {
						// GETNTHはSend
						msg.bmp = bmp;
						act.Reply(msg);
					}
					else if (msg.method == LoaderMessage.Method.FILL) {
						// 余計なキャッシュは削除。
						// 時間の掛かる処理じゃないのでbreak判定はない。
						// FILL指定時のみ。
						var kp = new List<int>();
						kp.Add(msg.index);
						for (int i = 0, n = msg.index + 1; i < POOL_COUNT + 3 && n < files.Count; i++, n++) {
							kp.Add(n);
						}
						for (int i = 0, n = msg.index - 1; i < POOL_COUNT + 3 && n >= 0; i++, n--) {
							kp.Add(n);
						}
						ImagesActKeep(kp);
                        if (!act.HasCue()) {
							if (zip != null) {
								zip.Dispose();
								zip = null;
							}
							if (fs != null) {
								fs.Dispose();
								fs = null;
								//Trace.WriteLine("BitmapPool Close Zip File");
							}
						}
					}
				}
			}
            if (zip != null) {
                zip.Dispose();
            }
            if (fs != null) {
                fs.Dispose();
            }
        }

        void ImagesActorMain(Object arg) {
            var act = (Actor)arg;
            // imagesはImagesActionを通してのみアクセスする、キャッシュの実体。
            Dictionary<int, Bitmap> images = new Dictionary<int, Bitmap>();
            Object o;
            while ((o = act.Receive()) != null) {
                var msg = (ImagesMessage)o;
                switch (msg.method) {
                    case ImagesMessage.Method.ADD:
                        images[msg.index] = msg.bmp;
                        break;
                    case ImagesMessage.Method.GETNTH:
                        msg.bmp = images.ContainsKey(msg.index) ? images[msg.index] : null;
						act.Reply(msg);
                        break;
                    case ImagesMessage.Method.KEEP:
                        var current = new List<int>();
                        foreach (int m in images.Keys) {
                            current.Add(m);
                        }
                        foreach (int m in current) {
                            if (!msg.keep.Contains(m)) {
                                images[m].Dispose();
                                images.Remove(m);
                            }
                        }
                        break;
                    case ImagesMessage.Method.CLEAR:
                        images.Clear();
                        act.Reply(null);
                        break;
                    default:
                        break;
                }
            }
        }
    }

    // XPのエクスプローラが自然順を導入してから結構経つのだが、いまだに.NETには採用されていない。
    // きっちりこのへんをサポートして.NETを盛り上げてくださいな。
    public sealed class NaturalOrderStringComparer : IComparer<string> {
		[DllImport("shlwapi.dll", CharSet = CharSet.Unicode)]
		private static extern int StrCmpLogicalW(string s1, string s2);
		//
		public int Compare(string s1, string s2) {
			return StrCmpLogicalW(s1, s2);
		}
	}
}
