﻿/*
	JumpDlg.xaml.cs
	Copyright 2016, OSTRA / Takeshi Yoneki
*/
using System;
using System.IO;
using System.Threading;
using System.Windows;
using System.Windows.Input;

namespace PictureSetViewLite {
    /// <summary>
    /// JumpDlg.xaml の相互作用ロジック
    /// </summary>
    public partial class JumpDlg : Window {
        public MainWindow mainView;
        // メインウィンドウからダイアログのページ番号を変えるとき、
        // 画像表示を変えないためのフラグ（既に変わっている結果として呼ばれるから）。
        bool enableJump = true;
        // スクロールコントロールの動きでメインウィンドウを変更したとき、
        // スクロールコントロールなどを変えないためのフラグ（変えてる原因はこっちなので）。
        // とはいえDispatcher.Invokeが絡むとそう簡単にはいかない。
        // 結局UIスレッドかどうか判断しなくちゃいけない。
        bool enableChangePageNumber = true;
        public Actor jumpAct = null;

		Actor pageAct = null;
		class PageMessage {
			public enum Method {
				NONE, GET, SET
			}
			public Method method = Method.NONE;
			public int page = 0;
		}

		public JumpDlg() {
            InitializeComponent();
            jumpAct = new Actor(JumpActorMain);
            pageAct = new Actor(PageActorMain);
        }

        public void ChangePageNumberText(int pn) {
            tbPageNumber.Text = (pn + 1).ToString();
            tbPageNumber.SelectAll();
        }

        int CalcPageNumber(int value) {
			if (!mainView.IsEnablePool()) {
				return 0;
			}
			return Prefs.data.ltor ? value : (mainView.pool.GetCount() - 1) - value;
        }

        // 初期化のときとメインウィンドウからダイアログのページ番号を変えるときに呼ぶ。
        // Jump処理はしない。
        public void ChangePageNumber(int pn) {
			enableJump = false;
			if (!scrlPage.IsMouseCaptureWithin && enableChangePageNumber) {
                ChangePageNumberText(pn);
                scrlPage.Value = CalcPageNumber(pn);
            }
			enableJump = true;
		}

        public void Setup() {
			if (!mainView.IsEnablePool()) {
				return;
			}
			scrlPage.Minimum = 0.0;
            scrlPage.Maximum = mainView.pool.GetCount() - 1;
            scrlPage.Value = CalcPageNumber(Prefs.data.page);
            scrlPage.ViewportSize = mainView.pool.GetCount() / 20.0;
            Title = Path.GetFileNameWithoutExtension(Prefs.data.filePath);
            ChangePageNumber(Prefs.data.page);
            btnBegin.Content = ((int)scrlPage.Minimum + 1).ToString();
            btnEnd.Content = ((int)scrlPage.Maximum + 1).ToString();
            tbPageNumber.Focus();
        }

        private void Jump(int pn, bool sync) {
			if (!mainView.IsEnablePool()) {
				return;
			}
            if (pn < 0) {
                pn = 0;
            }
            if (pn >= mainView.pool.GetCount()) {
                pn = mainView.pool.GetCount() - 1;
            }
            ChangePageNumberText(pn);
            if (sync && !jumpAct.IsWorking()) {
                // スレッドが動いているときは同期処理はあきらめる。
                JumpSub(pn, true);
            }
            else {
                pageAct.Post(new PageMessage { method = PageMessage.Method.SET, page = pn });
                // scrlPage_ValueChangedはものすごい量発生するのでここで梳く。
                if (!jumpAct.IsWorking()) {
                    // メッセージ自体はダミー
                    jumpAct.Post("");
                }
            }
        }

        public void Join() {
			if (jumpAct != null) {
				jumpAct.Quit();
				jumpAct.Join();
				jumpAct = null;
			}
			if (pageAct != null) {
				pageAct.Quit();
				pageAct.Join();
				pageAct = null;
			}
        }

        public void WaitActor() {
			// ActorがWaitになるまで待つ。
			while (jumpAct.IsWorking()) {
				Thread.Sleep(10);
			}
        }

        void JumpSub(int pn, bool bUiThread) {
            if (Prefs.data.page != pn) {
                Prefs.data.page = pn;
                mainView.pageType = PageType.NONE;
                mainView.DrawPicture(PageTurnRequest.CURRENT, bUiThread);
            }
        }

        void JumpActorMain(Object arg) {
            var act = (Actor)arg;
            while (act.Receive() != null) {
                enableChangePageNumber = false;
				var pnSaved = -1;
                while (true) {
                    Thread.Sleep(100);
					var pa = pageAct;
					if (pageAct == null) {
						break;
					}
                    int pn = (int)pa.Send(new PageMessage { method = PageMessage.Method.GET });
					if (pnSaved == pn) {
						break;
					}
					else { 
						pnSaved = pn;
						JumpSub(pn, false);
                    }
                }
                enableChangePageNumber = true;
            }
        }

        void PageActorMain(Object arg) {
            var act = (Actor)arg;
            int jumpPageNumber = 0;
            Object o;
            while ((o = act.Receive()) != null) {
				var msg = (PageMessage)o;
				switch (msg.method) {
					case PageMessage.Method.GET:
						act.Reply(jumpPageNumber);
						break;
					case PageMessage.Method.SET:
						jumpPageNumber = msg.page;
						break;
					default:
						break;
				}
            }
        }

        void Jump() {
            int pn = -1;
            try {
                pn = Int32.Parse(tbPageNumber.Text) - 1;
            }
            catch (Exception) {
            }
            Jump(pn, true);
        }

        private void btnOK_Click(object sender, RoutedEventArgs e) {
            Jump();
            Close();
        }

        private void Window_Closed(object sender, EventArgs e) {
            // 閉じた後のWindowは使いまわせないようだ。
            mainView.jumpDlg = null;
        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e) {
            // mainView.DrawPictureを呼んでるときにここに来てJoinしようとすると、デッドロックになる。
			if (jumpAct != null) {
                jumpAct.Quit();
				jumpAct = null;
			}
            if (pageAct != null) {
                pageAct.Quit();
				pageAct = null;
			}
        }

        private void btnClose_Click(object sender, RoutedEventArgs e) {
            Close();
        }

        private void scrlPage_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e) {
            int pn = CalcPageNumber((int)e.NewValue);
            if (enableJump) {
                Jump(pn, false);
            }
        }

        private void btnBegin_Click(object sender, RoutedEventArgs e) {
            ChangePageNumberText((int)scrlPage.Minimum);
			Jump();
			Close();
		}

		private void btnEnd_Click(object sender, RoutedEventArgs e) {
            ChangePageNumberText((int)scrlPage.Maximum);
			Jump();
			Close();
		}

		private void Window_KeyDown(object sender, System.Windows.Input.KeyEventArgs e) {
			if (e.Key == Key.Q) {
				Close();
				e.Handled = true;
			}
			else if (e.Key == Key.G) {
				mainView.Act();
				e.Handled = true;
			}
		}
	}
}
